@@ -1,7 +1,7 @@
 /*
  hstr.c     HSTR shell history completion utility
 
- Copyright (C) 2014-2018 Martin Dvorak <martin.dvorak@mindforger.com>
+ Copyright (C) 2014-2021 Martin Dvorak <martin.dvorak@mindforger.com>
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
@@ -67,28 +67,32 @@
 
 #define HSTR_ENV_VAR_CONFIG      "HSTR_CONFIG"
 #define HSTR_ENV_VAR_PROMPT      "HSTR_PROMPT"
+#define HSTR_ENV_VAR_IS_SUBSHELL "HSTR_IS_SUBSHELL"
 
-#define HSTR_CONFIG_THEME_MONOCHROMATIC   "monochromatic"
-#define HSTR_CONFIG_THEME_HICOLOR         "hicolor"
-#define HSTR_CONFIG_STATIC_FAVORITES      "static-favorites"
+#define HSTR_CONFIG_THEME_MONOCHROMATIC     "monochromatic"
+#define HSTR_CONFIG_THEME_HICOLOR           "hicolor"
+#define HSTR_CONFIG_STATIC_FAVORITES        "static-favorites"
 #define HSTR_CONFIG_SKIP_FAVORITES_COMMENTS "skip-favorites-comments"
-#define HSTR_CONFIG_FAVORITES         "favorites-view"
-#define HSTR_CONFIG_SORTING           "raw-history-view"
-#define HSTR_CONFIG_CASE              "case-sensitive"
-#define HSTR_CONFIG_REGEXP            "regexp-matching"
-#define HSTR_CONFIG_SUBSTRING         "substring-matching"
-#define HSTR_CONFIG_KEYWORDS          "keywords-matching"
-#define HSTR_CONFIG_NOCONFIRM         "no-confirm"
-#define HSTR_CONFIG_VERBOSE_KILL      "verbose-kill"
-#define HSTR_CONFIG_PROMPT_BOTTOM     "prompt-bottom"
-#define HSTR_CONFIG_BLACKLIST         "blacklist"
-#define HSTR_CONFIG_KEEP_PAGE         "keep-page"
-#define HSTR_CONFIG_DEBUG             "debug"
-#define HSTR_CONFIG_WARN              "warning"
-#define HSTR_CONFIG_BIG_KEYS_SKIP     "big-keys-skip"
-#define HSTR_CONFIG_BIG_KEYS_FLOOR    "big-keys-floor"
-#define HSTR_CONFIG_BIG_KEYS_EXIT     "big-keys-exit"
-#define HSTR_CONFIG_DUPLICATES        "duplicates"
+#define HSTR_CONFIG_FAVORITES               "favorites-view"
+#define HSTR_CONFIG_SORTING                 "raw-history-view"
+#define HSTR_CONFIG_CASE                    "case-sensitive"
+#define HSTR_CONFIG_REGEXP                  "regexp-matching"
+#define HSTR_CONFIG_SUBSTRING               "substring-matching"
+#define HSTR_CONFIG_KEYWORDS                "keywords-matching"
+#define HSTR_CONFIG_NO_CONFIRM              "no-confirm"
+#define HSTR_CONFIG_VERBOSE_KILL            "verbose-kill"
+#define HSTR_CONFIG_PROMPT_BOTTOM           "prompt-bottom"
+#define HSTR_CONFIG_HELP_ON_OPPOSITE_SIDE   "help-on-opposite-side"
+#define HSTR_CONFIG_HIDE_BASIC_HELP         "hide-basic-help"
+#define HSTR_CONFIG_HIDE_HELP               "hide-help"
+#define HSTR_CONFIG_BLACKLIST               "blacklist"
+#define HSTR_CONFIG_KEEP_PAGE               "keep-page"
+#define HSTR_CONFIG_DEBUG                   "debug"
+#define HSTR_CONFIG_WARN                    "warning"
+#define HSTR_CONFIG_BIG_KEYS_SKIP           "big-keys-skip"
+#define HSTR_CONFIG_BIG_KEYS_FLOOR          "big-keys-floor"
+#define HSTR_CONFIG_BIG_KEYS_EXIT           "big-keys-exit"
+#define HSTR_CONFIG_DUPLICATES              "duplicates"
 
 #define HSTR_DEBUG_LEVEL_NONE  0
 #define HSTR_DEBUG_LEVEL_WARN  1
@@ -133,7 +137,7 @@
 
 // major.minor.revision
 static const char* VERSION_STRING=
-        "hstr version \"2.1.0\" (2019-12-08T13:30:00)"
+        "hstr version \"2.4.0\" (2021-12-03T21:30:00)"
         "\n";
 
 static const char* HSTR_VIEW_LABELS[]={
@@ -171,7 +175,7 @@ static const char* INSTALL_BASH_STRING=
         //             It works correctly if memory entries are not deleted by HSTR. It doesn't synchronize history
         //             across different bash sessions.
         //   -c -r ... Forces entire .bash_history to be reloaded (handles history deletes, synchronizes different bash sessions)
-        "\n# ensure synchronization between Bash memory and history file"
+        "\n# ensure synchronization between bash memory and history file"
         "\nexport PROMPT_COMMAND=\"history -a; history -n; ${PROMPT_COMMAND}\""
         "\n# if this is interactive shell, then bind hstr to Ctrl-r (for Vi mode check doc)"
 #if defined(__MS_WSL__)
@@ -259,6 +263,7 @@ static const char* HELP_STRING=
         "\n  --show-configuration     -s ... show configuration to be added to ~/.bashrc"
         "\n  --show-zsh-configuration -z ... show zsh configuration to be added to ~/.zshrc"
         "\n  --show-blacklist         -b ... show commands to skip on history indexation"
+        "\n  --insert-in-terminal=[c] -i ... insert command c in terminal prompt and exit"
         "\n  --version                -V ... show version details"
         "\n  --help                   -h ... help"
         "\n"
@@ -275,15 +280,16 @@ static const char* LABEL_HELP=
 #define GETOPT_OPTIONAL_ARGUMENT     2
 
 static const struct option long_options[] = {
-        {"favorites",              GETOPT_NO_ARGUMENT, NULL, 'f'},
-        {"kill-last-command",      GETOPT_NO_ARGUMENT, NULL, 'k'},
-        {"version",                GETOPT_NO_ARGUMENT, NULL, 'V'},
-        {"help",                   GETOPT_NO_ARGUMENT, NULL, 'h'},
-        {"non-interactive",        GETOPT_NO_ARGUMENT, NULL, 'n'},
-        {"show-configuration",     GETOPT_NO_ARGUMENT, NULL, 's'},
-        {"show-zsh-configuration", GETOPT_NO_ARGUMENT, NULL, 'z'},
-        {"show-blacklist",         GETOPT_NO_ARGUMENT, NULL, 'b'},
-        {0,                        0,                  NULL,  0 }
+        {"favorites",              GETOPT_NO_ARGUMENT, NULL,       'f'},
+        {"kill-last-command",      GETOPT_NO_ARGUMENT, NULL,       'k'},
+        {"version",                GETOPT_NO_ARGUMENT, NULL,       'V'},
+        {"help",                   GETOPT_NO_ARGUMENT, NULL,       'h'},
+        {"non-interactive",        GETOPT_NO_ARGUMENT, NULL,       'n'},
+        {"show-configuration",     GETOPT_NO_ARGUMENT, NULL,       's'},
+        {"show-zsh-configuration", GETOPT_NO_ARGUMENT, NULL,       'z'},
+        {"show-blacklist",         GETOPT_NO_ARGUMENT, NULL,       'b'},
+        {"insert-in-terminal",     GETOPT_REQUIRED_ARGUMENT, NULL, 'i'},
+        {0,                        0,                  NULL,        0 }
 };
 
 typedef struct {
@@ -312,10 +318,14 @@ typedef struct {
     int bigKeys;
     int debugLevel;
     bool promptBottom;
+    bool helpOnOppositeSide;
+    bool hideBasicHelp;
+    bool hideHistoryHelp;
 
     int promptY;
     int promptYHelp;
     int promptYHistory;
+    int promptYNotification;
     int promptYItemsStart;
     int promptYItemsEnd;
     int promptItems;
@@ -353,10 +363,14 @@ void hstr_init(void)
     hstr->bigKeys=RADIX_BIG_KEYS_SKIP;
     hstr->debugLevel=HSTR_DEBUG_LEVEL_NONE;
     hstr->promptBottom=false;
+    hstr->helpOnOppositeSide=false;
+    hstr->hideBasicHelp=false;
+    hstr->hideHistoryHelp=false;
 
     hstr->promptY
      =hstr->promptYHelp
      =hstr->promptYHistory
+     =hstr->promptYNotification
      =hstr->promptYItemsStart
      =hstr->promptYItemsEnd
      =hstr->promptItems
@@ -377,36 +391,124 @@ void hstr_destroy(void)
     free(hstr);
 }
 
-void hstr_on_exit(void)
+void hstr_exit(int status)
 {
-    history_mgmt_flush();
     hstr_destroy();
+    exit(status);
 }
 
 void signal_callback_handler_ctrl_c(int signum)
 {
     if(signum==SIGINT) {
+        history_mgmt_flush();
         hstr_curses_stop(false);
-        hstr_on_exit();
-        exit(signum);
+        hstr_exit(signum);
     }
 }
 
 unsigned recalculate_max_history_items(void)
 {
-    hstr->promptItems = getmaxy(stdscr) - 3;
+    int n = getmaxy(stdscr);
+    hstr->promptItems = n-1;
+    if(!hstr->hideBasicHelp) {
+        hstr->promptItems--;
+    }
+    if(!hstr->hideHistoryHelp) {
+        hstr->promptItems--;
+    }
     if(hstr->promptBottom) {
-        hstr->promptY = getmaxy(stdscr) - 1;
-        hstr->promptYHelp = hstr->promptY - 1;
-        hstr->promptYHistory = hstr->promptY - 2;
-        hstr->promptYItemsStart = 0;
-        hstr->promptYItemsEnd = hstr->promptItems-1;
+        if(hstr->helpOnOppositeSide) {
+            // Layout:
+            // - [basic help]
+            // - [history help]
+            // - items start
+            // - ...
+            // - items end
+            // - prompt
+            int top = 0;
+            hstr->promptY = n-1;
+            if(!hstr->hideBasicHelp) {
+                hstr->promptYHelp = top++;
+            }
+            if(!hstr->hideHistoryHelp) {
+                hstr->promptYHistory = top++;
+            }
+            hstr->promptYItemsStart = top++;
+        } else {
+            // Layout:
+            // - items start
+            // - ...
+            // - items end
+            // - [history help]
+            // - [basic help]
+            // - prompt
+            int bottom = n-1;
+            hstr->promptY = bottom--;
+            if(!hstr->hideBasicHelp) {
+                hstr->promptYHelp = bottom--;
+            }
+            if(!hstr->hideHistoryHelp) {
+                hstr->promptYHistory = bottom--;
+            }
+            hstr->promptYItemsStart = 0;
+        }
     } else {
-        hstr->promptY = 0;
-        hstr->promptYHelp = 1;
-        hstr->promptYHistory = 2;
-        hstr->promptYItemsStart = 3;
-        hstr->promptYItemsEnd = getmaxy(stdscr);
+        if(hstr->helpOnOppositeSide) {
+            // Layout:
+            // - prompt
+            // - items start
+            // - ...
+            // - items end
+            // - [history help]
+            // - [basic help]
+            int bottom = n-1;
+            hstr->promptY = 0;
+            if(!hstr->hideBasicHelp) {
+                hstr->promptYHelp = bottom--;
+            }
+            if(!hstr->hideHistoryHelp) {
+                hstr->promptYHistory = bottom--;
+            }
+            hstr->promptYItemsStart = 1;
+        } else {
+            // Layout:
+            // - prompt
+            // - [basic help]
+            // - [history help]
+            // - items start
+            // - ...
+            // - items end
+            int top = 0;
+            hstr->promptY = top++;
+            if(!hstr->hideBasicHelp) {
+                hstr->promptYHelp = top++;
+            }
+            if(!hstr->hideHistoryHelp) {
+                hstr->promptYHistory = top++;
+            }
+            hstr->promptYItemsStart = top++;
+        }
+    }
+    hstr->promptYItemsEnd = hstr->promptYItemsStart+hstr->promptItems-1;
+    if(!hstr->hideBasicHelp) {
+        // Use basic help label for notifications.
+        hstr->promptYNotification = hstr->promptYHelp;
+    }
+    else {
+        // If basic help is hidden, we need another place to put notifications to.
+        if(hstr->hideBasicHelp) {
+            if(!hstr->hideHistoryHelp) {
+                // Use history help label.
+                hstr->promptYHelp = hstr->promptYHistory;
+            } else {
+                // Use one of the command item lines.
+                if((hstr->promptBottom && hstr->helpOnOppositeSide) || (!hstr->promptBottom && !hstr->helpOnOppositeSide)) {
+                    hstr->promptYHelp = hstr->promptYItemsStart;
+                } else {
+                    hstr->promptYHelp = hstr->promptYItemsEnd;
+                }
+            }
+        }
     }
     return hstr->promptItems;
 }
@@ -431,7 +533,7 @@ void hstr_get_env_configuration()
             if(strstr(hstr_config,HSTR_CONFIG_SUBSTRING)) {
                 hstr->matching=HSTR_MATCH_SUBSTRING;
             } else {
-                if(strstr(hstr_config, HSTR_CONFIG_KEYWORDS)) {
+                if(strstr(hstr_config,HSTR_CONFIG_KEYWORDS)) {
                     hstr->matching=HSTR_MATCH_KEYWORDS;
                 }
             }
@@ -461,7 +563,7 @@ void hstr_get_env_configuration()
         if(strstr(hstr_config,HSTR_CONFIG_KEEP_PAGE)) {
             hstr->keepPage=true;
         }
-        if(strstr(hstr_config,HSTR_CONFIG_NOCONFIRM)) {
+        if(strstr(hstr_config,HSTR_CONFIG_NO_CONFIRM)) {
             hstr->noConfirm=true;
         }
         if(strstr(hstr_config,HSTR_CONFIG_STATIC_FAVORITES)) {
@@ -488,6 +590,23 @@ void hstr_get_env_configuration()
         } else {
             hstr->promptBottom = false;
         }
+        if(strstr(hstr_config,HSTR_CONFIG_HELP_ON_OPPOSITE_SIDE)) {
+            hstr->helpOnOppositeSide = true;
+        } else {
+            hstr->helpOnOppositeSide = false;
+        }
+        if(strstr(hstr_config,HSTR_CONFIG_HIDE_HELP)) {
+            hstr->hideBasicHelp = true;
+            hstr->hideHistoryHelp = true;
+        } else {
+            if(strstr(hstr_config,HSTR_CONFIG_HIDE_BASIC_HELP)) {
+                hstr->hideBasicHelp = true;
+                hstr->hideHistoryHelp = false;
+            } else {
+                hstr->hideBasicHelp = false;
+                hstr->hideHistoryHelp = false;
+            }
+        }
         recalculate_max_history_items();
     }
 }
@@ -526,7 +645,7 @@ unsigned print_prompt(void)
 
 void add_to_selection(char* line, unsigned int* index)
 {
-    if (hstr->noRawHistoryDuplicates) {
+    if(hstr->noRawHistoryDuplicates) {
         unsigned i;
         for(i = 0; i < *index; i++) {
             if (strcmp(hstr->selection[i], line) == 0) {
@@ -535,11 +654,14 @@ void add_to_selection(char* line, unsigned int* index)
         }
     }
     hstr->selection[*index]=line;
-    *index = *index + 1;
+    (*index)++;
 }
 
 void print_help_label(void)
 {
+    if(hstr->hideBasicHelp)
+        return;
+
     int cursorX=getcurx(stdscr);
     int cursorY=getcury(stdscr);
 
@@ -554,13 +676,17 @@ void print_help_label(void)
 void print_confirm_delete(const char* cmd)
 {
     char screenLine[CMDLINE_LNG];
-    snprintf(screenLine, getmaxx(stdscr), "Do you want to delete all occurrences of '%s'? y/n", cmd);
+    if(hstr->view==HSTR_VIEW_FAVORITES) {
+        snprintf(screenLine, getmaxx(stdscr), "Do you want to delete favorites item '%s'? y/n", cmd);
+    } else {
+        snprintf(screenLine, getmaxx(stdscr), "Do you want to delete all occurrences of '%s'? y/n", cmd);
+    }
     // TODO make this function
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_on(COLOR_PAIR(HSTR_COLOR_DELETE));
         color_attr_on(A_BOLD);
     }
-    mvprintw(hstr->promptYHelp, 0, "%s", screenLine);
+    mvprintw(hstr->promptYNotification, 0, "%s", screenLine);
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_off(A_BOLD);
         color_attr_on(COLOR_PAIR(HSTR_COLOR_NORMAL));
@@ -572,13 +698,17 @@ void print_confirm_delete(const char* cmd)
 void print_cmd_deleted_label(const char* cmd, int occurences)
 {
     char screenLine[CMDLINE_LNG];
-    snprintf(screenLine, getmaxx(stdscr), "History item '%s' deleted (%d occurrence%s)", cmd, occurences, (occurences==1?"":"s"));
+    if(hstr->view==HSTR_VIEW_FAVORITES) {
+        snprintf(screenLine, getmaxx(stdscr), "Favorites item '%s' deleted", cmd);
+    } else {
+        snprintf(screenLine, getmaxx(stdscr), "History item '%s' deleted (%d occurrence%s)", cmd, occurences, (occurences==1?"":"s"));
+    }
     // TODO make this function
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_on(COLOR_PAIR(HSTR_COLOR_DELETE));
         color_attr_on(A_BOLD);
     }
-    mvprintw(hstr->promptYHelp, 0, "%s", screenLine);
+    mvprintw(hstr->promptYNotification, 0, "%s", screenLine);
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_off(A_BOLD);
         color_attr_on(COLOR_PAIR(HSTR_COLOR_NORMAL));
@@ -595,7 +725,7 @@ void print_regexp_error(const char* errorMessage)
         color_attr_on(COLOR_PAIR(HSTR_COLOR_DELETE));
         color_attr_on(A_BOLD);
     }
-    mvprintw(hstr->promptYHelp, 0, "%s", screenLine);
+    mvprintw(hstr->promptYNotification, 0, "%s", screenLine);
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_off(A_BOLD);
         color_attr_on(COLOR_PAIR(HSTR_COLOR_NORMAL));
@@ -612,7 +742,7 @@ void print_cmd_added_favorite_label(const char* cmd)
         color_attr_on(COLOR_PAIR(HSTR_COLOR_INFO));
         color_attr_on(A_BOLD);
     }
-    mvprintw(hstr->promptYHelp, 0, screenLine);
+    mvprintw(hstr->promptYNotification, 0, screenLine);
     if(hstr->theme & HSTR_THEME_COLOR) {
         color_attr_off(A_BOLD);
         color_attr_on(COLOR_PAIR(HSTR_COLOR_NORMAL));
@@ -623,6 +753,9 @@ void print_cmd_added_favorite_label(const char* cmd)
 
 void print_history_label(void)
 {
+    if(hstr->hideHistoryHelp)
+        return;
+
     unsigned width=getmaxx(stdscr);
 
     char screenLine[CMDLINE_LNG];
@@ -942,9 +1075,14 @@ char* hstr_print_selection(unsigned maxHistoryItems, char* pattern)
 
     move(hstr->promptYItemsStart, 0);
     clrtobot();
-    if(hstr->promptBottom) {
+    bool labelsAreOnBottom = (hstr->promptBottom && !hstr->helpOnOppositeSide) || (!hstr->promptBottom && hstr->helpOnOppositeSide);
+    if(labelsAreOnBottom) {
+        // TODO: Why is the reprinting here necessary? Please make a comment.
         print_help_label();
         print_history_label();
+    }
+    if(hstr->promptBottom) {
+        // TODO: Why is the reprinting here necessary? Please make a comment.
         print_pattern(pattern, hstr->promptY, print_prompt());
         y=hstr->promptYItemsEnd;
     } else {
@@ -1036,7 +1174,7 @@ void highlight_selection(int selectionCursorPosition, int previousSelectionCurso
 int remove_from_history_model(char* almostDead)
 {
     if(hstr->view==HSTR_VIEW_FAVORITES) {
-        return favorites_remove(hstr->favorites, almostDead);
+        return (int)favorites_remove(hstr->favorites, almostDead);
     } else {
         // raw & ranked history is pruned first as its items point to system history lines
         int systemOccurences=0, rawOccurences=history_mgmt_remove_from_raw(almostDead, hstr->history);
@@ -1071,17 +1209,37 @@ void stdout_history_and_return(void)
 // IMPROVE hstr doesn't have to be passed as parameter - it's global static
 char* getResultFromSelection(int selectionCursorPosition, Hstr* hstr, char* result) {
     if (hstr->promptBottom) {
-        result=hstr->selection[hstr->promptYItemsEnd-selectionCursorPosition];
+        result=hstr->selection[hstr->promptItems-selectionCursorPosition-1];
     } else {
         result=hstr->selection[selectionCursorPosition];
     }
     return result;
 }
 
+void hide_notification(void)
+{
+    if(!hstr->hideBasicHelp) {
+        print_help_label();
+    } else {
+        if(!hstr->hideHistoryHelp) {
+            print_history_label();
+        } else {
+            // TODO: If possible, we should rerender the command list here,
+            //  because one of the items was used to print the notification.
+        }
+    }
+}
+
 void loop_to_select(void)
 {
     signal(SIGINT, signal_callback_handler_ctrl_c);
 
+    bool isSubshellHint=FALSE;
+    char* isSubshellHintText = getenv(HSTR_ENV_VAR_IS_SUBSHELL);
+    if(isSubshellHintText && strlen(isSubshellHintText)>0) {
+        isSubshellHint=TRUE;
+    }
+
     hstr_curses_start();
     // TODO move the code below to hstr_curses
     color_init_pair(HSTR_COLOR_NORMAL, -1, -1);
@@ -1096,13 +1254,15 @@ void loop_to_select(void)
     // TODO why do I print non-filtered selection when on command line there is a pattern?
     hstr_print_selection(recalculate_max_history_items(), NULL);
     color_attr_off(COLOR_PAIR(HSTR_COLOR_NORMAL));
-    if(!hstr->promptBottom) {
+    bool labelsAreOnBottom = (hstr->promptBottom && !hstr->helpOnOppositeSide) || (!hstr->promptBottom && hstr->helpOnOppositeSide);
+    if(!labelsAreOnBottom) {
+        // TODO: Why is the reprinting here necessary? Please make a comment.
         print_help_label();
         print_history_label();
     }
 
     bool done=FALSE, skip=TRUE, executeResult=FALSE, lowercase=TRUE;
-    bool printDefaultLabel=TRUE, fixCommand=FALSE, editCommand=FALSE;
+    bool hideNotificationOnNextTick=TRUE, fixCommand=FALSE, editCommand=FALSE;
     unsigned basex=print_prompt();
     int x=basex, c, cc, cursorX=0, cursorY=0, maxHistoryItems, deletedOccurences;
     int width=getmaxx(stdscr);
@@ -1134,9 +1294,9 @@ void loop_to_select(void)
             continue;
         }
 
-        if(printDefaultLabel) {
-            print_help_label();
-            printDefaultLabel=FALSE;
+        if(hideNotificationOnNextTick) {
+            hide_notification();
+            hideNotificationOnNextTick=FALSE;
         }
 
         if(c == K_CTRL_R) {
@@ -1165,12 +1325,12 @@ void loop_to_select(void)
                     result=hstr_print_selection(maxHistoryItems, pattern);
                     print_cmd_deleted_label(msg, deletedOccurences);
                 } else {
-                    print_help_label();
+                    hide_notification();
                 }
                 free(msg);
                 move(hstr->promptY, basex+strlen(pattern));
-                printDefaultLabel=TRUE;
-                print_history_label();
+                hideNotificationOnNextTick=TRUE;
+                print_history_label();  // TODO: Why is this necessary? Add comment!
 
                 if(hstr->selectionSize == 0) {
                     // just update the cursor, there are no elements to select
@@ -1179,12 +1339,12 @@ void loop_to_select(void)
                 }
 
                 if(hstr->promptBottom) {
-                    if(selectionCursorPosition <= (int)(hstr->promptYItemsEnd-hstr->selectionSize+1)) {
-                        selectionCursorPosition=hstr->promptYItemsEnd-hstr->selectionSize+1;
+                    if(selectionCursorPosition < (int)(hstr->promptItems-hstr->selectionSize)) {
+                        selectionCursorPosition=hstr->promptItems-hstr->selectionSize;
                     }
                 } else {
                     if(selectionCursorPosition >= (int)hstr->selectionSize) {
-                        selectionCursorPosition = hstr->selectionSize - 1;
+                        selectionCursorPosition = hstr->selectionSize-1;
                     }
                 }
                 highlight_selection(selectionCursorPosition, SELECTION_CURSOR_IN_PROMPT, pattern);
@@ -1244,7 +1404,7 @@ void loop_to_select(void)
                 selectionCursorPosition=SELECTION_CURSOR_IN_PROMPT;
                 if(hstr->view!=HSTR_VIEW_FAVORITES) {
                     print_cmd_added_favorite_label(result);
-                    printDefaultLabel=TRUE;
+                    hideNotificationOnNextTick=TRUE;
                 }
                 // TODO code review
                 if(strlen(pattern)<(width-basex-1)) {
@@ -1300,8 +1460,8 @@ void loop_to_select(void)
             previousSelectionCursorPosition=selectionCursorPosition;
             if(selectionCursorPosition>0) {
                 if(hstr->promptBottom) {
-                    if(selectionCursorPosition <= (int)(hstr->promptYItemsEnd-hstr->selectionSize+1)) {
-                        selectionCursorPosition=hstr->promptYItemsEnd;
+                    if(selectionCursorPosition <= (int)(hstr->promptItems-hstr->selectionSize)) {
+                        selectionCursorPosition=hstr->promptItems-1;
                     } else {
                         selectionCursorPosition--;
                     }
@@ -1310,7 +1470,7 @@ void loop_to_select(void)
                 }
             } else {
                 if(hstr->promptBottom) {
-                    selectionCursorPosition=hstr->promptYItemsEnd;
+                    selectionCursorPosition=hstr->promptItems-1;
                 } else {
                     selectionCursorPosition=hstr->selectionSize-1;
                 }
@@ -1333,17 +1493,17 @@ void loop_to_select(void)
         case K_CTRL_N:
             if(selectionCursorPosition==SELECTION_CURSOR_IN_PROMPT) {
                 if(hstr->promptBottom) {
-                    selectionCursorPosition=hstr->promptYItemsEnd-hstr->selectionSize+1;
+                    selectionCursorPosition=hstr->promptItems-hstr->selectionSize;
                 } else {
                     selectionCursorPosition=previousSelectionCursorPosition=0;
                 }
             } else {
                 previousSelectionCursorPosition=selectionCursorPosition;
                 if(hstr->promptBottom) {
-                    if(selectionCursorPosition<hstr->promptYItemsEnd) {
+                    if(selectionCursorPosition<hstr->promptItems-1) {
                         selectionCursorPosition++;
                     } else {
-                        selectionCursorPosition=hstr->promptYItemsEnd-hstr->selectionSize+1;
+                        selectionCursorPosition=hstr->promptItems-hstr->selectionSize;
                     }
                 } else {
                     if((selectionCursorPosition+1) < (int)hstr->selectionSize) {
@@ -1405,7 +1565,12 @@ void loop_to_select(void)
             break;
         case K_TAB:
         case KEY_RIGHT:
-            editCommand=TRUE;
+            if(!isSubshellHint) {
+                editCommand=TRUE;
+            } else {
+                // Not setting editCommand to TRUE here,
+                // because else an unnecessary blank line gets emitted before returning to prompt.
+            }
             if(selectionCursorPosition!=SELECTION_CURSOR_IN_PROMPT) {
                 result=getResultFromSelection(selectionCursorPosition, hstr, result);
                 if(hstr->view==HSTR_VIEW_FAVORITES) {
@@ -1419,7 +1584,7 @@ void loop_to_select(void)
         case K_CTRL_G:
         case K_ESC:
             result=NULL;
-            history_clear_dirty();
+            history_mgmt_clear_dirty();
             done=TRUE;
             break;
         case K_CTRL_X:
@@ -1496,16 +1661,16 @@ void hstr_interactive(void)
         } else {
             stdout_history_and_return();
         }
-        hstr_on_exit();
-    } else {
-        printf("No history - nothing to suggest...\n");
-    }
+        history_mgmt_flush();
+    } // else (no history) handled in create() method
+
+    hstr_exit(EXIT_SUCCESS);
 }
 
 void hstr_getopt(int argc, char **argv)
 {
     int option_index = 0;
-    int option = getopt_long(argc, argv, "fkVhnszb", long_options, &option_index);
+    int option = getopt_long(argc, argv, "fkVhnszbi", long_options, &option_index);
     if(option != -1) {
         switch(option) {
         case 'f':
@@ -1516,23 +1681,33 @@ void hstr_getopt(int argc, char **argv)
             break;
         case 'k':
             if(history_mgmt_remove_last_history_entry(hstr->verboseKill)) {
-                exit(EXIT_SUCCESS);
+                hstr_exit(EXIT_SUCCESS);
+                break;
             } else {
-                exit(EXIT_FAILURE);
+                hstr_exit(EXIT_FAILURE);
+                break;
             }
         case 'b':
             blacklist_load(&hstr->blacklist);
             blacklist_dump(&hstr->blacklist);
-            exit(EXIT_SUCCESS);
+            hstr_exit(EXIT_SUCCESS);
+            break;
+        case 'i':
+            fill_terminal_input(optarg, FALSE);
+            hstr_exit(EXIT_SUCCESS);
+            break;
         case 'V':
             printf("%s", VERSION_STRING);
-            exit(EXIT_SUCCESS);
+            hstr_exit(EXIT_SUCCESS);
+            break;
         case 'h':
             printf("%s", HELP_STRING);
-            exit(EXIT_SUCCESS);
+            hstr_exit(EXIT_SUCCESS);
+            break;
         case 'z':
             printf("%s", INSTALL_ZSH_STRING);
-            exit(EXIT_SUCCESS);
+            hstr_exit(EXIT_SUCCESS);
+            break;
         case 's':
             // ZSH_VERSION is not exported by zsh > detected by parent process name
             if(isZshParentShell()) {
@@ -1540,12 +1715,12 @@ void hstr_getopt(int argc, char **argv)
             } else {
                 printf("%s", INSTALL_BASH_STRING);
             }
-            exit(EXIT_SUCCESS);
-
+            hstr_exit(EXIT_SUCCESS);
+            break;
         case '?':
         default:
             printf("%s", HELP_STRING);
-            exit(EXIT_SUCCESS);
+            hstr_exit(EXIT_SUCCESS);
         }
     }
 
@@ -1565,9 +1740,8 @@ int hstr_main(int argc, char* argv[])
     hstr_getopt(argc, argv);
     favorites_get(hstr->favorites);
     blacklist_load(&hstr->blacklist);
+    // hstr cleanup is handled by hstr_exit()
     hstr_interactive();
 
-    // hstr cleanup is handled by hstr_on_exit()
-
     return EXIT_SUCCESS;
 }
